package org.kantega.jexmec.lifecycle;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


import org.kantega.jexmec.ClassLoaderProvider;
import org.kantega.jexmec.PluginLoader;
import org.kantega.jexmec.PluginManager;
import org.kantega.jexmec.PluginManagerListener;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;

/**
 * This listener provides life cycle support for plugins.
 * Which actual annotations will trigger events specified by the user
 * calling the various <code>set*Annotation</code> methods. By
 * default no annotations are configured, and thus no events are propagated to plugins.
 */
public class AnnotationBasedPluginLifecyleListener<P> extends PluginManagerListener<P> {

    private Class<? extends Annotation> beforeActivationAnnotation = null;
    private Class<? extends Annotation> afterActivationAnnotation = null;
    private Class<? extends Annotation> beforePassivationAnnotation = null;
    private Class<? extends Annotation> afterPassivationAnnotation = null;
    private Class<? extends Annotation> pluginsUpdatedAnnotation = null;


    public AnnotationBasedPluginLifecyleListener<P> setBeforeActivationAnnotation(Class<? extends Annotation> beforeActivationAnnotation) {
        this.beforeActivationAnnotation = beforeActivationAnnotation;
        return this;
    }

    public AnnotationBasedPluginLifecyleListener<P> setAfterActivationAnnotation(Class<? extends Annotation> afterActivationAnnotation) {
        this.afterActivationAnnotation = afterActivationAnnotation;
        return this;
    }

    public AnnotationBasedPluginLifecyleListener<P> setBeforePassivationAnnotation(Class<? extends Annotation> beforePassivationAnnotation) {
        this.beforePassivationAnnotation = beforePassivationAnnotation;
        return this;
    }

    public AnnotationBasedPluginLifecyleListener<P> setAfterPassivationAnnotation(Class<? extends Annotation> afterPassivationAnnotation) {
        this.afterPassivationAnnotation = afterPassivationAnnotation;
        return this;
    }

    public AnnotationBasedPluginLifecyleListener<P> setPluginsUpdatedAnnotation(Class<? extends Annotation> pluginsUpdatedAnnotation) {
        this.pluginsUpdatedAnnotation = pluginsUpdatedAnnotation;
        return this;
    }



    /**
     * For all plugins to be activated, invoke any method annotated with whatever annotation is configured by {@link #setBeforeActivationAnnotation(Class)}
     */
    @Override
    public void beforeActivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pPluginLoader, Collection<P> plugins) {
        runLifeCyclePhase(beforeActivationAnnotation, plugins);
    }

    /**
     * For all activated plugins, invoke any method annotated with whatever annotation is configured by {@link #setAfterActivationAnnotation(Class)}
     */
    @Override
    public void afterActivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pPluginLoader, Collection<P> plugins) {
        runLifeCyclePhase(afterActivationAnnotation, plugins);
    }

    /**
     * For all plugins to be passivated, invoke any method annotated with whatever annotation is configured by {@link #setBeforePassivationAnnotation(Class)}
     */
    @Override
    public void beforePassivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pPluginLoader, Collection<P> plugins) {
        runLifeCyclePhase(beforePassivationAnnotation, plugins);
    }

    /**
     * For all passivated plugins, invoke any method annotated with whatever annotation is configured by {@link #setAfterPassivationAnnotation(Class)}
     */
    @Override
    public void afterPassivation(PluginManager<P> pluginManager, ClassLoaderProvider classLoaderProvider, ClassLoader classLoader, PluginLoader<P> pluginLoader, Collection<P> plugins) {
        runLifeCyclePhase(afterPassivationAnnotation, plugins);
    }

    /**
     * For all plugins active in the registry, invoke any method annotated with whatever annotation configured by {@link #setPluginsUpdatedAnnotation(Class)}
     * @param plugins the currently in the registry
     */
    @Override
    public void pluginsUpdated(Collection<P> plugins) {
        runLifeCyclePhase(pluginsUpdatedAnnotation, plugins,  plugins);
    }

    private void runLifeCyclePhase(Class<? extends Annotation> phase, Iterable<P> plugins, Object... parameters) {
        if(phase != null) {
            for(P plugin : plugins) {
                for(Method method : plugin.getClass().getDeclaredMethods()) {
                    if(method.getAnnotation(phase) != null) {
                        validateLifecycleMethod(method, parameters, beforeActivationAnnotation);

                        try {
                            method.invoke(plugin, parameters);
                        } catch (IllegalAccessException e) {
                            exceptionInvokingLifecycleMethod(phase, plugin, method, e);
                        } catch (InvocationTargetException e) {
                            exceptionInvokingLifecycleMethod(phase, plugin, method, e);
                        }
                    }
                }
            }
        }
    }

    /**
     * Validates that an annotated method returns <code>void</code> and that it can accept the provided parameters.
     * @param method the plugin method to validate
     * @param parameters any parameters that this method is expected to take
     * @param annotation the annotation which triggered invocation of this method.
     */
    private void validateLifecycleMethod(Method method, Object[] parameters, Class<? extends Annotation> annotation) {
        if(method.getReturnType() != void.class) {
            invalidPluginMethod(method, annotation, " The method does not return void as expected, but instead returns " + method.getReturnType());
        }

        Class[] paramTypes = new Class[parameters.length];
        for (int i = 0; i < paramTypes.length; i++) {
            paramTypes[i] = parameters[i] == null ? null : parameters[i].getClass();

        }
        String invalidParameterListMessage = "The method takes parameters " + Arrays.toString(method.getParameterTypes())
                + (parameters.length > 0 ? " Expected parameters " + Arrays.toString(paramTypes) : " Expected no parameters")
                + ". ";
        if(method.getParameterTypes().length != parameters.length) {
            invalidPluginMethod(method, annotation, invalidParameterListMessage);
        }

        for (int i = 0; i < method.getParameterTypes().length; i++) {
            if(parameters[i] != null && !method.getParameterTypes()[i].isAssignableFrom(paramTypes[i])) {
                invalidPluginMethod(method, annotation, "Actual parameter " + parameters[i].getClass() + " is not assignable to parameter type" + method.getParameterTypes()[i]);
            }

        }
    }

    protected void exceptionInvokingLifecycleMethod(Class<? extends Annotation> phase, P plugin, Method method, Throwable e) {
        throw new RuntimeException("Exception invoking life cycle method " + method.getName() + " for phase " + phase.getName() +" on plugin class " + plugin.getClass().getName(), e);
    }

    protected void invalidPluginMethod(Method method, Class<? extends Annotation> annotation, String reason) {
        throw new RuntimeException("Method " + method.getName() + " of plugin class " + method.getDeclaringClass().getName() +" is annotated with "
        + annotation.getName() + "." + reason);
    }
}
